//! Encapsulated logic components.
pub mod apply_extra_ops;
pub mod bridge;
pub mod column_selector;
pub mod default_field_values;
pub mod errors;
pub mod forbid_extra_columns;
pub mod physical_slice;
pub mod projection;
pub mod reader_operation_pushdown;
pub mod row_counter;
pub mod row_deletions;
